"""Checker for grammar errors."""

import json
import re
import subprocess
import tempfile
from pathlib import Path
from typing import Set
from zipfile import ZipFile

import requests
import simplelogging

from padpo.checkers.baseclass import Checker, replace_quotes
from padpo.pofile import PoItem, PoFile

log = simplelogging.get_logger()


class GrammalecteChecker(Checker):
    """Checker for grammar errors."""

    name = "Grammalecte"

    def __init__(self):
        """Initialiser."""
        super().__init__()
        self.dir = None
        self.personal_dict: Set[str] = set()
        self.get_personal_dict()

    @staticmethod
    def run_grammalecte(filename: str) -> subprocess.CompletedProcess:
        return subprocess.run(
            [
                "grammalecte-cli.py",
                "-f",
                filename,
                "-off",
                "apos",
                "--json",
                "--only_when_errors",
            ],
            capture_output=True,
            text=True,
        )

    def check_file(self, pofile: PoFile):
        """Check a `*.po` file."""
        if not isinstance(pofile, PoFile):
            log.error("%s is not an instance of PoFile", str(pofile))
        _, filename = tempfile.mkstemp(
            suffix=".txt", prefix="padpo_", text=True
        )
        with open(filename, "w", encoding="utf8") as f:
            text = pofile.rst2txt()
            text = re.sub(r"«\s(.*?)\s»", replace_quotes, text)
            f.write(text)
        try:
            result = self.run_grammalecte(filename)
        except FileNotFoundError as e:
            if e.filename == "grammalecte-cli.py":
                install_grammalecte()
                result = self.run_grammalecte(filename)
        if result.stdout:
            warnings = json.loads(result.stdout)
            self.manage_grammar_errors(warnings, pofile)
            self.manage_spelling_errors(warnings, pofile)
        Path(filename).unlink()

    def check_item(self, item: PoItem):
        """Check an item in a `*.po` file (does nothing)."""
        pass

    def manage_grammar_errors(self, warnings, pofile: PoFile):
        """Manage grammar errors returned by grammalecte."""
        for warning in warnings["data"]:
            for error in warning["lGrammarErrors"]:
                if self.filter_out_grammar_error(error):
                    continue
                item_index = int(warning["iParagraph"]) // 2
                item = pofile.content[item_index]
                start = max(0, int(error["nStart"]) - 40)
                end = max(0, int(error["nEnd"]) + 10)
                item.add_warning(
                    self.name,
                    f'{error["sMessage"]} => '
                    f"###{item.msgstr_rst2txt[start:end]}###",
                )

    def filter_out_grammar_error(self, error):
        """Return True when grammalecte error should be ignored."""
        msg = error["sRuleId"]
        if msg in (
            "esp_milieu_ligne",  # double space
            "nbsp_avant_deux_points",  # NBSP
            "nbsp_avant_double_ponctuation",  # NBSP
        ):
            return True
        if "typo_guillemets_typographiques_simples" in msg:
            return True  # ignore ' quotes
        msg_text = error["sMessage"]
        if msg_text in (
            "Accord de genre erroné : « ABC » est masculin.",
            "Accord de genre erroné : « PEP » est masculin.",
            "Accord de nombre erroné : « PEP » devrait être au pluriel.",
            "Accord de genre erroné : « une entrée » est féminin, « utilisateur » est masculin.",
        ):
            return True
        if "S’il s’agit d’un impératif" in msg_text:
            if error["nStart"] == 0:
                # ignore imperative conjugation at begining of 1st sentence
                return True
        return False

    def manage_spelling_errors(self, warnings, pofile: PoFile):
        """Manage spelling errors returned by grammalecte."""
        for warning in warnings["data"]:
            for error in warning["lSpellingErrors"]:
                if self.filter_out_spelling_error(error):
                    continue
                item_index = int(warning["iParagraph"]) // 2
                item = pofile.content[item_index]
                start = max(0, int(error["nStart"]) - 40)
                end = max(0, int(error["nEnd"]) + 10)
                word = error["sValue"]
                item.add_warning(
                    self.name,
                    f'Unknown word "{word}" in '
                    f"###{item.msgstr_rst2txt[start:end]}###",
                )

    def filter_out_spelling_error(self, error):
        """Return True when grammalecte error should be ignored."""
        word = error["sValue"]
        if set(word) == {"x"}:
            return True  # word is xxxxx or xxxxxxxx…
        if word.strip() in self.personal_dict:
            return True  # white list
        return False

    def get_personal_dict(self):
        """
        Add spelling white list.

        Based on
        https://raw.githubusercontent.com/python/python-docs-fr/3.8/dict
        """
        download_request = requests.get(
            "https://raw.githubusercontent.com/python/python-docs-fr/3.8/dict"
        )
        download_request.raise_for_status()
        for line in download_request.text.splitlines():
            word = line.strip()
            self.personal_dict.add(word)
            self.personal_dict.add(word.title())


def install_grammalecte():
    """Install grammalecte CLI."""
    log.warning("Missing grammalecte, trying to install it")
    tmpdirname = tempfile.mkdtemp(prefix="padpo_grammalecte_")
    tmpdirname = Path(tmpdirname)
    tmpdirname.mkdir(exist_ok=True)
    download_request = requests.get(
        "https://grammalecte.net/grammalecte/zip/Grammalecte-fr-v1.5.0.zip"
    )
    download_request.raise_for_status()
    zip_file = tmpdirname / "Grammalecte-fr-v1.5.0.zip"
    zip_file.write_bytes(download_request.content)
    with ZipFile(zip_file, "r") as zip_obj:
        zip_obj.extractall(tmpdirname / "Grammalecte-fr-v1.5.0")
    subprocess.run(
        ["pip", "install", str(tmpdirname / "Grammalecte-fr-v1.5.0")]
    )



"""Checker for missing non breakable spaces."""

import re

from padpo.checkers.baseclass import Checker, replace_quotes
from padpo.pofile import PoItem


class NonBreakableSpaceChecker(Checker):
    """Checker for missing non breakable spaces."""

    name = "NBSP"

    def check_item(self, item: PoItem):
        """Check an item in a `*.po` file."""
        text = item.msgstr_rst2txt
        for match in re.finditer(r"(.{0,30})(«[^ ])(.{0,30})", text):
            self.__add_message(item, *match.groups())
        for match in re.finditer(r"(.{0,30})([^ ][»])(.{0,30})", text):
            self.__add_message(item, *match.groups())
        text = re.sub(r"«\s(.*?)\s»", replace_quotes, text)
        text = re.sub(r"http://", "http-//", text)
        text = re.sub(r"https://", "https-//", text)
        for sign in "?!:;":
            regex = r"(.{0,30})([^ ][" + sign + r"])(.{0,30})"
            for match in re.finditer(regex, text):
                prefix = item.msgstr_rst2txt[match.start(1) : match.end(1)]
                suffix = item.msgstr_rst2txt[match.start(3) : match.end(3)]
                match = item.msgstr_rst2txt[match.start(2) : match.end(2)]
                self.__add_message_space_before(item, prefix, match, suffix)

    def __add_message(self, item, prefix, match, suffix):
        item.add_error(
            self.name,
            "Space should be replaced with a non-breakable space in "
            f'"{match}": between ###{prefix}### and ###{suffix}###',
        )

    def __add_message_space_before(self, item, prefix, match, suffix):
        item.add_error(
            self.name,
            f"There should be a non-breakable space before "
            f'"{match[1:]}": between ###{prefix}{match[0]}### and '
            f"###{match[1:]}{suffix}###",
        )



"""Checker for double spaces."""

import re

from padpo.checkers.baseclass import Checker
from padpo.pofile import PoItem


class DoubleSpaceChecker(Checker):
    """Checker for double spaces."""

    name = "Double space"

    def check_item(self, item: PoItem):
        """Check an item in a `*.po` file."""
        for match in re.finditer(
            r"(.{0,30})\s\s(.{0,30})", item.msgstr_full_content
        ):
            item.add_warning(
                self.name,
                f"Double spaces detected between ###{match.group(1)}### "
                f"and ###{match.group(2)}###",
            )



"""Entry point of padpo."""

import argparse
import sys
from pathlib import Path

import simplelogging

from padpo.pofile import PoFile
from padpo.checkers import checkers
from padpo.github import pull_request_files


log = None


def check_file(path, pull_request_info=None):
    """Check a `*.po` file."""
    pofile = PoFile(path)

    for checker in checkers:
        checker.check_file(pofile)

    return pofile.display_warnings(pull_request_info)


def check_directory(path, pull_request_info=None):
    """Check a directory containing `*.po` files."""
    path = Path(path)
    result_errors = []
    result_warnings = []
    for filepath in path.rglob("*.po"):
        errors, warnings = check_file(filepath, pull_request_info)
        result_errors.extend(errors)
        result_warnings.extend(warnings)
    return result_errors, result_warnings


def check_path(path, pull_request_info=None):
    """Check a path (`*.po` file or directory)."""
    path = Path(path)
    if path.is_dir():
        return check_directory(path, pull_request_info)
    else:
        return check_file(path, pull_request_info)


def main():
    """Entry point."""
    global log
    log = simplelogging.get_logger("__main__")

    parser = argparse.ArgumentParser(description="Linter for *.po files.")
    parser.add_argument("-v", "--verbose", action="count", default=0)
    files = parser.add_mutually_exclusive_group(required=True)
    files.add_argument(
        "-i",
        "--input-path",
        metavar="PATH",
        type=str,
        help="path of the file or directory to check",
        default="",
    )
    files.add_argument(
        "-g",
        "--github",
        metavar="python/python-docs-fr/pull/978",
        type=str,
        help="path of pull request in GitHub to check",
        default="",
    )
    files.add_argument(
        "-p",
        "--python-docs-fr",
        metavar="978",
        type=int,
        help="ID of pull request in python-docs-fr repository",
        default=0,
    )
    args = parser.parse_args()
    if args.verbose < 1:
        log.reduced_logging()
    elif args.verbose < 2:
        log.normal_logging()
    else:
        log.full_logging()

    if args.input_path:
        path = args.input_path
        pull_request_info = None
    else:
        pull_request = ""
        if args.github:
            pull_request = args.github
        if args.python_docs_fr:
            pull_request = f"python/python-docs-fr/pull/{args.python_docs_fr}"
        pull_request_info = pull_request_files(pull_request)
        path = pull_request_info.download_directory

    errors, warnings = check_path(path, pull_request_info=pull_request_info)
    if errors:
        sys.exit(1)



"""Checker for missing translations."""


from padpo.checkers.baseclass import Checker
from padpo.pofile import PoItem


class EmptyChecker(Checker):
    """Checker for missing translations."""

    name = "Empty"

    def check_item(self, item: PoItem):
        """Check an item in a `*.po` file."""
        if item.msgid_full_content and not item.msgstr_full_content:
            item.add_warning(self.name, "This entry is not translated yet.")



"""Checkers list."""
from padpo.checkers.doublespace import DoubleSpaceChecker
from padpo.checkers.empty import EmptyChecker
from padpo.checkers.fuzzy import FuzzyChecker
from padpo.checkers.grammalecte import GrammalecteChecker
from padpo.checkers.linelength import LineLengthChecker
from padpo.checkers.nbsp import NonBreakableSpaceChecker

checkers = [
    DoubleSpaceChecker(),
    EmptyChecker(),
    FuzzyChecker(),
    GrammalecteChecker(),
    LineLengthChecker(),
    NonBreakableSpaceChecker(),
]



"""Base class for checkers."""

from abc import ABC, abstractmethod

import simplelogging

from padpo.pofile import PoFile, PoItem

log = simplelogging.get_logger()


class Checker(ABC):
    """Base class for checkers."""

    name = "UnknownChecker"  # name displayed in error messages

    def check_file(self, pofile: PoFile):
        """Check a `*.po` file."""
        if not isinstance(pofile, PoFile):
            log.error("%s is not an instance of PoFile", str(pofile))
        for item in pofile.content:
            self.check_item(item)

    @abstractmethod
    def check_item(self, item: PoItem):
        """Check an item in a `*.po` file."""
        return NotImplementedError


def replace_quotes(match):
    """Replace match with « xxxxxxx »."""
    length = len(match.group(0)) - 4
    return "« " + length * "x" + " »"



"""Checker for fuzzy translations."""

from padpo.checkers.baseclass import Checker
from padpo.pofile import PoItem


class FuzzyChecker(Checker):
    """Checker for fuzzy translations."""

    name = "Fuzzy"

    def check_item(self, item: PoItem):
        """Check an item in a `*.po` file."""
        if item.fuzzy:
            item.add_warning(self.name, "This entry is tagged as fuzzy.")



"""Managment of `*.po` files."""

import re
from typing import List

import simplelogging

log = simplelogging.get_logger()


class PoItem:
    """Translation item."""

    def __init__(self, path, lineno):
        """Initializer."""
        self.path = path[3:]
        self.lineno_start = lineno
        self.lineno_end = lineno
        self.parsing_msgid = None
        self.msgid = []
        self.msgstr = []
        self.fuzzy = False
        self.warnings = []
        self.inside_pull_request = False

    def append_line(self, line):
        """Append a line of a `*.po` file to the item."""
        self.lineno_end += 1
        if line.startswith("msgid"):
            self.parsing_msgid = True
            self.msgid.append(line[7:-2])
        elif line.startswith("msgstr"):
            self.parsing_msgid = False
            self.msgstr.append(line[8:-2])
        elif line.startswith("#, fuzzy"):
            self.fuzzy = True
        elif line.startswith('"'):
            if self.parsing_msgid:
                self.msgid.append(line[1:-2])
            elif not self.parsing_msgid is None:
                self.msgstr.append(line[1:-2])

    def __str__(self):
        """Return string representation."""
        return (
            f"    - {self.msgid_full_content}\n"
            f"        => {self.msgstr_full_content}\n"
            f"        => {self.msgstr_rst2txt}\n"
        )

    @property
    def msgid_full_content(self):
        """Full content of the msgid."""
        return "".join(self.msgid)

    @property
    def msgstr_full_content(self):
        """Full content of the msgstr."""
        return "".join(self.msgstr)

    @property
    def msgid_rst2txt(self):
        """Full content of the msgid (reStructuredText escaped)."""
        return self.rst2txt(self.msgid_full_content)

    @property
    def msgstr_rst2txt(self):
        """Full content of the msgstr (reStructuredText escaped)."""
        return self.rst2txt(self.msgstr_full_content)

    @staticmethod
    def rst2txt(text):
        """
        Escape reStructuredText markup.

        The text is modified to transform reStructuredText markup
        in textual version. For instance:

        * "::" becomes ":"
        * ":class:`PoFile`" becomes "« PoFile »"
        """
        text = re.sub(r"::", r":", text)
        text = re.sub(r"``(.*?)``", r"« \1 »", text)
        text = re.sub(r"\"(.*?)\"", r"« \1 »", text)
        text = re.sub(r":pep:`(.*?)`", r"PEP \1", text)
        text = re.sub(r":[a-z:]+:`(.+?)`", r"« \1 »", text)
        text = re.sub(r"\*\*(.*?)\*\*", r"« \1 »", text)
        text = re.sub(
            r"\*(.*?)\*", r"« \1 »", text
        )  # TODO sauf si déjà entre «»
        text = re.sub(
            r"`(.*?)\s*<((?:http|https|ftp)://.*?)>`_", r"\1 (« \2 »)", text
        )
        text = re.sub(r"<((?:http|https|ftp)://.*?)>", r"« \1 »", text)
        return text

    def add_warning(self, checker_name: str, text: str) -> None:
        """Add a checker warning to the item."""
        self.warnings.append(Warning(checker_name, text))

    def add_error(self, checker_name: str, text: str) -> None:
        """Add a checker error to the item."""
        self.warnings.append(Error(checker_name, text))


class PoFile:
    """A `*.po` file information."""

    def __init__(self, path=None):
        """Initializer."""
        self.content: List[PoItem] = []
        self.path = path
        if path:
            self.parse_file(path)

    def parse_file(self, path):
        """Parse a `*.po` file according to its path."""
        # TODO assert path is a file, not a dir
        item = None
        with open(path, encoding="utf8") as f:
            for lineno, line in enumerate(f):
                if line.startswith("#: "):
                    if item:
                        self.content.append(item)
                    item = PoItem(line, lineno + 1)
                elif item:
                    item.append_line(line)
        if item:
            self.content.append(item)

    def __str__(self):
        """Return string representation."""
        ret = f"Po file: {self.path}\n"
        ret += "\n".join(str(item) for item in self.content)
        return ret

    def rst2txt(self):
        """Escape reStructuredText markup."""
        return "\n\n".join(item.msgstr_rst2txt for item in self.content)

    def display_warnings(self, pull_request_info=None):
        """Log warnings and errors, return errors and warnings lists."""
        self.tag_in_pull_request(pull_request_info)
        errors = []
        warnings = []
        for item in self.content:
            if not item.inside_pull_request:
                continue
            prefix = f"{self.path}:{item.lineno_start:-4} %s"
            log.debug(prefix, "")
            for message in item.warnings:
                if isinstance(message, Error):
                    log.error(prefix, message)
                    errors.append(message)
                elif isinstance(message, Warning):
                    log.warning(prefix, message)
                    warnings.append(message)
        return errors, warnings

    def tag_in_pull_request(self, pull_request_info):
        """Tag items being part of the pull request."""
        if not pull_request_info:
            for item in self.content:
                item.inside_pull_request = True
        else:
            diff = pull_request_info.diff(self.path)
            for item in self.content:
                item.inside_pull_request = False
            for lineno_diff in self.lines_in_diff(diff):
                for item in self.content:
                    if item.lineno_start <= lineno_diff <= item.lineno_end:
                        item.inside_pull_request = True

    @staticmethod
    def lines_in_diff(diff):
        """Yield line numbers modified in a diff (new line numbers)."""
        for line in diff.splitlines():
            if line.startswith("@@"):
                match = re.search(r"@@\s*\-\d+,\d+\s+\+(\d+),(\d+)\s+@@", line)
                if match:
                    line_start = int(match.group(1))
                    nb_lines = int(match.group(2))
                    # github add 3 extra lines around diff info
                    extra_info_lines = 3
                    for lineno in range(
                        line_start + extra_info_lines,
                        line_start + nb_lines - extra_info_lines,
                    ):
                        yield lineno


class Message:
    """Checker message."""

    def __init__(self, checker_name: str, text: str):
        """Initializer."""
        self.checker_name = checker_name
        self.text = text

    def __str__(self):
        """Return string representation."""
        return f"[{self.checker_name:^14}] {self.text}"


class Warning(Message):
    """Checker warning message."""


class Error(Message):
    """Checker error message."""



"""GitHub interactions."""

import tempfile
from pathlib import Path

import requests


class PullRequestInfo:
    """Information on a pull request."""

    def __init__(self):
        """Initializer."""
        self._data = {}
        self.download_directory = None

    def add_file(self, filename, temp_path, diff):
        """Add file info to the pull request."""
        self._data[str(temp_path)] = (temp_path, diff, filename)

    def diff(self, path):
        """Return diff of a file in the pull request."""
        if str(path) in self._data:
            return self._data[str(path)][1]
        return ""

    def temp_path(self, path):
        """Return temporary file path of the file in the pull request."""
        if str(path) in self._data:
            return self._data[str(path)][0]
        return ""

    def filename(self, path):
        """Return file name of a file in the pull request."""
        if str(path) in self._data:
            return self._data[str(path)][2]
        return ""


def pull_request_files(pull_request: str):
    """Return pull request information."""
    pull_request = pull_request.replace("/pull/", "/pulls/")
    request = requests.get(
        f"https://api.github.com/repos/{pull_request}/files"
    )
    request.raise_for_status()
    # TODO remove directory at end of execution
    temp_dir = tempfile.mkdtemp(prefix="padpo_")
    pr = PullRequestInfo()
    pr.download_directory = temp_dir
    for fileinfo in request.json():
        filename = fileinfo["filename"]
        temp_file = Path(temp_dir) / filename
        content_request = requests.get(fileinfo["raw_url"])
        content_request.raise_for_status()
        temp_file_dir = temp_file.parent
        temp_file_dir.mkdir(parents=True, exist_ok=True)
        temp_file.write_bytes(content_request.content)
        pr.add_file(filename, temp_file, fileinfo["patch"])
    return pr



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['padpo', 'padpo.checkers']

package_data = \
{'': ['*']}

install_requires = \
['requests==2.22.0',
 'setuptools==42.0.2',
 'simplelogging==0.10.0',
 'wheel==0.33.6']

entry_points = \
{'console_scripts': ['padpo = padpo.padpo:main']}

setup_kwargs = {
    'name': 'padpo',
    'version': '0.5.0',
    'description': 'Linter for gettext files',
    'long_description': "# padpo\n\nLinter for gettext files (*.po)\n\nCreated to help the translation of official Python docs in French: https://github.com/python/python-docs-fr\n\nIl faut demander aux traducteurs s'ils n'ont pas de pot quand ils traduisent, maintenant ils ont `padpo`…\n:smile: :laughing: :stuck_out_tongue_winking_eye: :joy: (note\xa0: il était tard le soir quand j'ai trouvé le nom).\n\n**WORK IN PROGRESS**\n\n## License\n\nBSD 3-clause\n\nPull request are welcome.\n\n## Usage\n\nUsing the *activated virtual environment* created during the installation:\n\nFor a local input file:\n\n```bash\npadpo --input-path a_file.po\n```\n\nor for a local input directory:\n\n```bash\npadpo --input-path a_directory_containing_po_files\n```\n\nor for a pull request in python-docs-fr repository (here pull request #978)\n\n```bash\npadpo --python-docs-fr 978\n```\n\nor for a pull request in a GitHub repository (here python/python-docs-fr/pull/978)\n\n```bash\npadpo --github python/python-docs-fr/pull/978\n```\n\n![Screenshot](screenshot.png)\n\n## Installation\n\n### Automatic installation\n\n```bash\npip install padpo\n```\n\n### Manual installation\n\n1. Create a virtual environment with Python 3.7 and above\n\n   ```bash\n   python3.7 -m venv venv\n   ```\n\n2. Activate the virtual environment\n\n   ```bash\n   source venv/bin/activate\n   ```\n\n3. Install dependencies\n\n   ```bash\n   poetry install\n   ```\n\n   Note: this uses `poetry` that you can get here: https://poetry.eustace.io/docs/\n\n## Update on PyPI\n\n`./deliver.sh`\n\n## Changelog\n\n### v0.5.0 (2019-12-3)\n\n* check spelling errors with grammalecte\n* tag releases!\n\n### v0.4.0 (2019-12-2)\n\n* use poetry: https://poetry.eustace.io/docs/\n* add some tests with tox and pytests\n* fix some false positive issues with grammalecte\n",
    'author': 'Vincent Poulailleau',
    'author_email': 'vpoulailleau@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/vpoulailleau/padpo',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""Checker for line length."""

from padpo.checkers.baseclass import Checker
from padpo.pofile import PoItem

MAX_LINE_LENGTH = 79


class LineLengthChecker(Checker):
    """Checker for line length."""

    name = "Line length"

    def check_item(self, item: PoItem):
        """Check an item in a `*.po` file."""
        for line in item.msgstr:
            if len(line) > MAX_LINE_LENGTH - 2:  # 2 is for ""
                item.add_error(
                    self.name,
                    f"Line too long ({len(line) + 2} > "
                    f"{MAX_LINE_LENGTH}): {line}",
                )
